      SUBROUTINE DECOMP(M, N, NSTEP, T0, H, XQ, CVAL, F, E, LDE, A, LDA,
     &                  T, LDT, EL, LDEL, TU,
     &                  LDTU, Q1, LDQ1, Q2, LDQ2, Q3, LDQ3, TAU1, TAU2,
     &                  TAU3, RWORK, LRWORK, EDIF, ADIF,
     &                  FDIF, IPAR, RPAR, MAXIND, IOPT)
C
C  Purpose:
C 
C   This subroutine computes the solution of the minimization problem
C                            .
C      ||x||_2=min!  s.t. ||Ex-Ax-f||_2=min!,
C
C   where
C       .
C      Ex = Ax + f, x(t_0)=0 
C
C   is a strangeness-free linear differential algebraic equation in
C   orthogonal standard form. This solution is computed by solving
C   the minimization problem
C                            
C      ||x_h||_2=min!  s.t. ||D_h*x_h-f_h||_2=min!
C
C   where the linear system
C
C      D_h*x_h = f_h
C
C   represents the discretization of the above differential algebraic
C   system with the implicit Euler method and with a homogegeous 
C   initial value. This system is solved with respect to the 
C   block bidiagonal structure of the matrix D_h.
C
C   The parameters are supplied by the subroutine DGLOBAL.
C
C     .. Scalar Arguments..
      INTEGER          M, N, NSTEP, LDE, LDA, LDT, LDTU, LDQ1, LDQ2, 
     &                 LDQ3, LRWORK
      DOUBLE PRECISION T0, H
C     .. Array Arguments ..
      INTEGER          CVAL(*), IPAR(*), IOPT(*)
      DOUBLE PRECISION XQ(*), F(*), E(LDE,*), A(LDA,*), T(LDT,*), 
     &                 EL(LDEL,*), TU(LDTU,*), 
     &                 Q1(LDQ1,*), Q2(LDQ2,*), Q3(LDQ3,*), TAU1(*),
     &                 TAU2(*), TAU3(*), RWORK(*), RPAR(*)
C     .. Subroutine Arguments ..
      EXTERNAL         EDIF, ADIF, FDIF 
C     .. Local Scalars ..
      INTEGER          I, J, I2
      DOUBLE PRECISION TI 
C     .. Local Scalars ..
      INTEGER          CVAL2(5)
C     .. Executable Statements ..
C
C   Initialize F=h*f_N
C
      IF (IOPT(1) .EQ. 0) THEN
         NEQ = CVAL(2) + CVAL(3)
	ELSE
	   NEQ = M
	END IF
      CALL DSCAL(NEQ, H, F, 1)
C
C   Backward loop: decomposition of the discretization matrix
C    
      DO 200 I2=NSTEP,2,-1
	   CALL DLASET('A', NEQ, 2*NEQ, 0.0D0, 0.0D0, EL, LDEL)	  
         IPOSM = (I2-1)*NEQ+1
         IPOSN = (I2-1)*N+1
C
C   Compute LQ-decomposition T = [L 0]*Q_{i1} and set EL = [* L] 
C    
         CALL DGELQF(NEQ, N, T, LDT, TAU1(IPOSM), RWORK, LRWORK, IERR)
	   CALL DLACPY('L', NEQ, NEQ, T, LDT, EL(1,NEQ+1), LDEL)
	   CALL DLACPY('U', NEQ, N, T, LDT, Q1(IPOSM,1), LDQ1)
C
C   Compute LQ-decomposition -E = [\tilde{E} L]*Q_{i2} and set 
C   EL = [\tilde{E} L]
C
         CALL DGELQF(NEQ, N, E, LDE, TAU2(IPOSM), RWORK, LRWORK, IERR)
	   CALL DLACPY('L', NEQ, NEQ, E, LDE, EL, LDEL)
	   CALL DLACPY('U', NEQ, N, E, LDE, Q2(IPOSM,1), LDQ2)
C
C   Update T = (E_{i-1} - h*A_{i-1})*Q_{2i}^T, transform to full rank
C   if necessary and set TU = [T(:,1:NEQ) 0]
C     
         TI=T0+(I2-1)*H
	   IF (IOPT(1) .EQ. 0) THEN
	      CALL REDUCE(EDIF, ADIF, FDIF, M, N, TI, MAXIND, CVAL2, E,
     $                  LDE, A, LDA, F(M+1), IPAR, RPAR, RWORK, LRWORK,
     $                  IERR)
	   ELSE
	      CALL EDIF(M, N, TI, 0, E, LDE, IPAR, RPAR, IERR)
	      CALL ADIF(M, N, TI, 0, A, LDA, IPAR, RPAR, IERR)
	      CALL FDIF(M, TI, 0, F(M+1), IPAR, RPAR, IERR)
	   END IF
	   IF (CVAL2(1) .NE. CVAL(1) .OR. CVAL2(2) .NE. CVAL(2) .OR.
     $       CVAL2(3) .NE. CVAL(3) .OR. CVAL2(4) .NE. CVAL(4) .OR.
     $       CVAL2(5) .NE. CVAL(5)) THEN
	      WRITE (*,*) 'Error in subroutine DECOMP at time T=:',TI
	      WRITE (*,*) 'Characteristic values have changed'
	   END IF
	   DO 210 I=1,NEQ
	      DO 210 J=1,N
	         T(I,J) = E(I,J) - H*A(I,J)
	         E(I,J) = -E(I,J)
 210     CONTINUE
         CALL DORMLQ('R','T', NEQ, N, NEQ, Q2(IPOSM,1), LDQ2,
     &               TAU2(IPOSM), T, LDT, RWORK, LRWORK, IERR)
         CALL DLACPY('A', NEQ, NEQ, T, LDT, TU, LDTU)
	   CALL DLASET('A', NEQ, NEQ, 0.0D0, 0.0D0, TU(1,NEQ+1), 
     &               LDTU)
C	  
C   Compute RQ-decomposition EL = [\tilde E  L]*Q_{i3}
C
         CALL DGERQF(NEQ, 2*NEQ, EL, LDEL, TAU3(IPOSM), RWORK, 
     &               LRWORK, IERR)
	   CALL DLACPY('A', NEQ, 2*NEQ, EL, LDEL, Q3(IPOSM,1), LDQ3)
C
C   Update [T(:,1:NEQ) U] = [T(:,1:NEQ) 0]*Q_3^T
C 
    	   CALL DORMRQ('R','T', NEQ, 2*NEQ, NEQ, Q3(IPOSM,1), LDQ3,
     &               TAU3(IPOSM), TU, LDTU, RWORK, LRWORK, IERR)
         CALL DLACPY('A', NEQ, NEQ, TU, LDTU, T, LDT)
C
C   Compute XQ((i-1)*N+1:(i-1)*N+M) = R^(-1)*F
C	 
	   CALL DTRTRS('U','N','N', NEQ, 1, Q3(IPOSM,NEQ+1), LDQ3,
     &               F, NEQ*NSTEP, IERR)
         CALL DCOPY(NEQ, F, 1, XQ(IPOSN), 1)
         CALL DLASET('A', N-NEQ, 1, 0.0D0, 0.0D0, XQ(IPOSN+NEQ), M)
C
C   Update F = f_{i-1} - U*XQ((i-1)*N+1:(i-1)*N+M)
C
         CALL DCOPY(NEQ, F(M+1), 1, F, 1)
         CALL DSCAL(NEQ, H, F, 1)
	   CALL DGEMV('N', NEQ, NEQ, -1.0D0, TU(1,NEQ+1), LDTU,
     &              XQ(IPOSN), 1, 1.0D0, F, 1)	   
 200  CONTINUE 
C
C   Compute RQ-decomposition T = [0 R]Q_{11}
C
      CALL DGERQF(NEQ, N, T, LDT, TAU1, RWORK, LRWORK, IERR)
C
C   Compute XQ(N-NEQ+1:N) = R^(-1)*F
C     
      CALL DTRTRS('U','N','N', NEQ, 1, T(1,N-NEQ+1), LDT, F,
     &            M*NSTEP, IERR)
      CALL DCOPY(NEQ, F, 1, RWORK(N-NEQ+1), 1)
      CALL DLASET('A', N-NEQ, 1, 0.0D0, 0.0D0, RWORK, M)
      CALL DCOPY(N, RWORK, 1, XQ, 1)
C
C   Compute XQ(1:N) = Q_1^T*XQ(1:N)
C   
      CALL DORMRQ('L','T', N, 1, NEQ, T, LDT, TAU1, XQ, NSTEP*N, 
     &            RWORK, LRWORK, IERR)
C
C   Forward loop: compute the solution
C
      DO 300 I2=2,NSTEP
         IPOSM = (I2-1)*NEQ+1
         IPOSN = (I2-1)*N+1
	   IPOSN2= (I2-2)*N+1
C
C   Update [XQ((i-2)N+1:(i-2)N+M)] = Q_{i3}^T*[XQ((i-2)N+1:(i-2)N+M)]
C          [XQ((i-1)N+1:(i-1)N+M)]            [XQ((i-1)N+1:(i-1)N+M)]
C
         CALL DCOPY(NEQ, XQ(IPOSN2), 1, RWORK, 1)
	   CALL DCOPY(NEQ, XQ(IPOSN), 1, RWORK(NEQ+1), 1)
	   CALL DORMRQ('L','T', 2*NEQ, 1, NEQ, Q3(IPOSM,1), LDQ3,
     &               TAU3(IPOSM), RWORK, NSTEP*M, RWORK(2*NEQ+1), 
     &               LRWORK, IERR)
         CALL DCOPY(NEQ, RWORK, 1, XQ(IPOSN2), 1)
	   CALL DCOPY(NEQ, RWORK(NEQ+1), 1, XQ(IPOSN), 1)
C
C   Update XQ((i-2)N+1:(i-1)N) = Q_{i2}^T*XQ((i-2)N+1:(i-1)N)
C
	   CALL DORMLQ('L','T', N, 1, NEQ, Q2(IPOSM,1), LDQ2,
     &               TAU2(IPOSM), XQ(IPOSN2), NSTEP*N, RWORK, LRWORK,
     &               IERR)
C
C   Update XQ((i-1)N+1:iN) = Q_{i2}^T*XQ((i-1)N+1:iN)
C
	   CALL DORMLQ('L','T', N, 1, NEQ, Q1(IPOSM,1), LDQ1,
     &               TAU1(IPOSM), XQ(IPOSN), NSTEP*N, RWORK, LRWORK,
     &               IERR)
	 
 300  CONTINUE 
      RETURN
      END
