      SUBROUTINE SLVIVP(EDEF, ADEF, FDEF, NSTEP, XQ, RWORK, LRWK, 
     $                  IWORK, LIWK)
C     .. Scalar Arguments ..
	INTEGER          NSTEP, LRWK, LIWK
C     .. Array Arguments ..
      INTEGER          IWORK(*)
	DOUBLE PRECISION XQ(*), RWORK(*)
C     .. Subroutine Arguments
      EXTERNAL         EDEF, ADEF, FDEF
C
C   Purpose:
C
C   Computes the solution of the initial value problem
C                 .
C     [\Sigma_E 0]x=[A_{11}  A_{12} ]x + [f_1]
C     [   0     0]  [A_{21} \Sigma_A]    [f_2]
C
C   which is taken as a reference solution for the approximations
C   computed by the subroutine DLMIN. The solution is computed with
C   the linear DAE-solver DGELDA.
C
C     .. Local Scalars ..
	INTEGER          M, N, ID, IA, IU, IV, NVAR, OX, OXPRIME, ORPAR,
     $                 ORWORK, LRWORK, METHOD
	DOUBLE PRECISION T0, TN, TOUT, H, RTOL, ATOL
C     .. Local Arrays ..
	INTEGER          IPAR(1), CVAL(4), INFO(20)
C     .. External Functions ..
	EXTERNAL         SETSIZE, SETINT, INIVAL, EDIF, ADIF, FDIF
C     .. Executable Statements ..
C
C   Get the properties of this problem.
C
      CALL SETSIZE(M, N, ID, IA, IU, IV)
	NVAR = ID + IA
	CALL SETINT(T0, TN)
      H = (TN - T0)/NSTEP
C
C   Compute consistent initial values
C
      IPAR(1) = 0
	CALL INIVAL(NVAR, ID, IA, T0, XQ, IPAR, RWORK)
C
C   Setup RWORK-array and parameters for DGELDA
C
	OX      = 1
	OXPRIME = OX      + NVAR
	ORPAR   = OXPRIME + NVAR
	ORWORK  = ORPAR   + M*N + 5*(M+N)
	LRWORK  = LRWK - ORWORK
	IF (LRWORK .LT. 50+26*NVAR+42*NVAR**2) THEN
	   WRITE (*,*) 'SLVIVP: RWORK-array too small for DGELDA'
	   WRITE (*,*) 'LRWORK should be at least', 50+26*NVAR+42*NVAR**2
     $                                            +ORWORK
	END IF
	IF (LIWK .LT. 20+2*NVAR) THEN
	   WRITE (*,*) 'SLVIVP: IWORK-array too small for DGELDA'
	   WRITE (*,*) 'LIWK should be at least', 20+2*NVAR
	END IF
C
C   Set tolerances and options for DGELDA:
C   
C   The absolute and relative error-tolerance are set to 1.0D-10,
C   the maximum strangeness index is set to zero since the system is
C   strangeness-free, and a consistent initial value is provided.
C
	RTOL = 1.0D-10
	ATOL = 1.0D-10
	METHOD = 1
	DO 10 I = 1,20
	   INFO(I) = 0
  10  CONTINUE
      INFO(10) = 1
	IWORK(4) = 0
	INFO(11) = 1
	CALL DCOPY(NVAR, XQ, 1, RWORK(OX), 1)
C
C   Compute reference solution at NSTEP intermediate points
C
	DO 100 I = 1,NSTEP
	   TOUT = I*H
	   CALL DGELDA(EDIF, ADIF, FDIF, NVAR, T0, TOUT, RWORK(OX), 
     $               RWORK(OXPRIME), CVAL, IPAR, RWORK(ORPAR), IWORK,
     $               LIWK, RWORK(ORWORK), LRWORK, RTOL, ATOL, METHOD,
     $               INFO, IWARN, IERR)
         IF (IERR.LT.0) THEN
            CALL DGEERM(EDIF, ADIF, FDIF, NEQ, T, TOUT, X, XPRIME, CVAL,
     $        IPAR, RWORK(ORPAR), IWORK, LIWK, RWORK, LRWORK,
     $        RTOL, ATOL, METHOD, INFO, IWARN, IERR)
            STOP
         ENDIF
	   CALL DCOPY(NVAR, RWORK(OX), 1, XQ((I-1)*N+1),1)
	   DO 50 J = NVAR+1,N
	      XQ((I-1)*N+J) = 0.0D0
  50     CONTINUE
 100  CONTINUE
	RETURN
	END

	SUBROUTINE EDIF(N,T,IDIF,E,LDE,IPAR,RPAR,IERR)
C     .. Scalar Arguments ..
      DOUBLE PRECISION T
      INTEGER          IDIF, IERR, LDE, N
C     .. Array Arguments ..
      DOUBLE PRECISION E(LDE,*), RPAR(*)
	INTEGER          IPAR(*)
C
C   This is the subroutine that defines the coefficient matrix E for the
C   reference solution as used by the code DGELDA. Furthermore it is 
C   checked if the subroutine EDEF is set up correctly, i.e. that the
C   matrix E defined by EDEF corresponds with a system in orthogonal
C   standard form.
C
C     .. Local Scalars ..
      INTEGER          M, N2, ID, IA, IU, IV, OS, OU, OV, ORWRK, LRWRK
C     .. Local Arrays ..
      DOUBLE PRECISION RWORK(1000)
C     .. Executable Statements ..
      CALL SETSIZE(M, N2, ID, IA, IU, IV)
	CALL EDEF(M, N2, T, IDIF, RPAR, M, IPAR, RPAR, IERR)
	IF (IPAR(1) .EQ. 0) THEN
         DO 10 I = 1,ID
	      DO 10 J = ID, N2-1
	         IF (RPAR(M*J+I) .NE. 0.0D0) THEN
	            WRITE (*,*) 'Error: E not given in standard form'
	            WRITE (*,*) 'Program terminated'
	            STOP
	         END IF
  10     CONTINUE
         DO 20 I = ID+1,M
	      DO 20 J = 0, N2-1
	         IF (RPAR(M*J+I) .NE. 0.0D0) THEN
	            WRITE (*,*) 'Error: E not given in standard form'
	            WRITE (*,*) 'Program terminated'
	            STOP
	         END IF
  20     CONTINUE
      END IF
	CALL DLACPY('A', N, N, RPAR, M, E, LDE)
	OS    = 1
	OU    = OS + ID
	OV    = OU + 1
	ORWRK = OV + 1
	LRWRK = 1000 - ORWRK
	CALL DGESVD('N', 'N', ID, ID, E, M, RPAR(OS), RPAR(OU), 1,
     $            RPAR(OV), 1, RPAR(ORWRK), LRWRK, IERR)
	IF (RPAR(OS+ID-1) .LE. 1.0D-12) THEN
	   WRITE (*,*) 'Error: rank of Sigma_E < d'
	   WRITE (*,*) 'Program terminated'
	   STOP
	END IF
	RETURN
	END
      

	SUBROUTINE ADIF(N,T,IDIF,A,LDA,IPAR,RPAR,IERR)
C     .. Scalar Arguments ..
      DOUBLE PRECISION T
      INTEGER          IDIF, IERR, LDA, N
C     .. Array Arguments ..
      DOUBLE PRECISION A(LDA,*), RPAR(*)
	INTEGER          IPAR(*)
C
C   This is the subroutine that defines the coefficient matrix A for the
C   reference solution as used by the code DGELDA. Furthermore it is 
C   checked if the subroutine ADEF is set up correctly, i.e. that the
C   matrix A defined by ADEF corresponds with a system in orthogonal
C   standard form.
C
C     .. Local Scalars ..
      INTEGER          M, N2, ID, IA, IU, IV, OS, OU, OV, ORWRK, LRWRK
C     .. Local Arrays ..
      DOUBLE PRECISION RWORK(1000)
C     .. Executable Statements ..
      CALL SETSIZE(M, N2, ID, IA, IU, IV)
	CALL ADEF(M, N2, T, IDIF, RPAR, M, IPAR, RPAR, IERR)
	IF (IPAR(1) .EQ. 0) THEN
         DO 10 I = ID+1,ID+IA
	      DO 10 J = ID+IA, N2-1
	         IF (RPAR(M*J+I) .NE. 0.0D0) THEN
	            WRITE (*,*) 'Error: A not given in standard form'
	            WRITE (*,*) 'Program terminated'
	            STOP
	         END IF
  10     CONTINUE
         DO 20 I = ID+IA+1,M
	      DO 20 J = 0, N2-1
	         IF (RPAR(M*J+I) .NE. 0.0D0) THEN
	            WRITE (*,*) 'Error: A not given in standard form'
	            WRITE (*,*) 'Program terminated'
	            STOP
	         END IF
  20     CONTINUE
      END IF
	CALL DLACPY('A', N, N, RPAR, M, A, LDA)
	IF (IA .GT. 0) THEN
	   OS    = 1
	   OU    = OS + ID
	   OV    = OU + 1
	   ORWRK = OV + 1
	   LRWRK = 1000 - ORWRK
	   CALL DGESVD('N', 'N', IA, IA, A(ID+1,ID+1), M, RPAR(OS),
     $               RPAR(OU), 1, RPAR(OV), 1, RPAR(ORWRK), LRWRK, 
     $               IERR)
	   IF (RPAR(OS+IA-1) .LE. 1.0D-12) THEN
	      WRITE (*,*) 'Error: rank of Sigma_A < a'
	      WRITE (*,*) 'Program terminated'
	      STOP
	   END IF
	END IF
	RETURN
	END
      

	SUBROUTINE FDIF(N,T,IDIF,F,IPAR,RPAR,IERR)
C     .. Scalar Arguments ..
      DOUBLE PRECISION T
      INTEGER          IDIF, IERR, N
C     .. Array Arguments ..
      DOUBLE PRECISION F(*), RPAR(*)
	INTEGER          IPAR(*)
C
C   This subroutine sets up the inhomogeneity for the computation of
C   the reference solution with DGENDA.
C
C     .. Local Scalars ..
      INTEGER          M, N2, ID, IA, IU, IV
C     .. Local Arrays ..
      DOUBLE PRECISION RWORK(1000)
C     .. Executable Statements ..
      CALL SETSIZE(M, N2, ID, IA, IU, IV)
	CALL FDEF(M, T, IDIF, RPAR, IPAR, RPAR, IERR)
	CALL DCOPY(N, RPAR, 1, F, 1)
	RETURN
	END


      SUBROUTINE INIVAL(N, ID, IA, T0, X, IPAR, RPAR)
C     .. Subroutine Arguments ..
      EXTERNAL         EDIF, ADIF, FDIF
C     .. Scalar Arguments ..
      INTEGER          N, ID, IA
	DOUBLE PRECISION T0
C
C   Computes the consistent initial value for the system in orthogonal
C   standard form by setting the differential part x_{10}=x_1(t_0) of 
C   initial value to zero and solving the equation
C     0=A_{11}(t_0)x_{10}+\Sigma_A(t_0)x_{20}+f(t_0)
C   with respect to the algebraic component x_{20}=x_2(t_0).
C
C     .. Array Arguments ..
      DOUBLE PRECISION X(*), RPAR(*)
	INTEGER          IPAR(*)
	DOUBLE PRECISION A(N,N), F(N)
	INTEGER          IPIV(IA)
	DO 10 I = 1,ID
	   X(I) = 0.0D0
  10  CONTINUE
      IF (IA .GT. 0) THEN
	   CALL ADIF(N, T0, 0, A, 2, 0, RPAR, IERR)
	   CALL FDIF(N, T0, 0, F, 0, RPAR, IERR)
	   DO 20 I = ID+1,ID+IA
	      X(I) = -F(I)
  20     CONTINUE
         CALL DGESV(IA, 1, A(ID+1,ID+1), N, IPIV, X(ID+1), IA, IERR)
	END IF
	RETURN
	END
